"""Test the multilevel sensor command class constants."""

from zwave_js_server.const.command_class.multilevel_sensor import (
    UNIT_ABSOLUTE_HUMIDITY,
    UNIT_HERTZ,
    UNIT_KILOGRAM,
    UNIT_MOLE_PER_CUBIC_METER,
    UNIT_PERCENTAGE_VALUE,
)


def test_multilevel_sensor_constants():
    """Test some of the multilevel sensor command class constants.

    Ensure that the enums are not overwriting each other.
    """
    assert len(UNIT_ABSOLUTE_HUMIDITY) == 1
    assert len(UNIT_HERTZ) == 2
    assert len(UNIT_KILOGRAM) == 2
    assert len(UNIT_MOLE_PER_CUBIC_METER) == 7
    assert len(UNIT_PERCENTAGE_VALUE) == 6



"""Test Energy Production command class utility functions."""

import pytest

from zwave_js_server.const import CommandClass
from zwave_js_server.const.command_class.energy_production import (
    CC_SPECIFIC_PARAMETER,
    CC_SPECIFIC_SCALE,
    EnergyProductionParameter,
    PowerScale,
)
from zwave_js_server.exceptions import InvalidCommandClass, UnknownValueData
from zwave_js_server.model.node import Node
from zwave_js_server.model.value import (
    MetaDataType,
    Value,
    ValueDataType,
    get_value_id_str,
)
from zwave_js_server.util.command_class.energy_production import (
    get_energy_production_parameter,
    get_energy_production_scale_type,
)


async def test_get_energy_production_parameter(energy_production: Node):
    """Test get_energy_production_parameter function."""
    node = energy_production

    value_id = get_value_id_str(node, CommandClass.VERSION, "protocolVersion")
    with pytest.raises(InvalidCommandClass):
        get_energy_production_parameter(node.values.get(value_id))

    value_id = get_value_id_str(
        node, CommandClass.ENERGY_PRODUCTION, "value", property_key=0
    )
    assert (
        get_energy_production_parameter(node.values.get(value_id))
        == EnergyProductionParameter.POWER
    )


async def test_invalid_get_energy_production_parameter(
    invalid_multilevel_sensor_type: Node,
):
    """Test receiving an invalid energy production parameter."""
    node = invalid_multilevel_sensor_type

    # Create value with an invalid parameter ID
    metadata = MetaDataType(ccSpecific={CC_SPECIFIC_PARAMETER: -1})
    value_data = ValueDataType(
        commandClass=CommandClass.ENERGY_PRODUCTION,
        commandClassName="EnergyProduction",
        endpoint=0,
        property="value",
        propertyKey=0,
        metadata=metadata,
    )
    value = Value(node, value_data)
    with pytest.raises(UnknownValueData):
        get_energy_production_parameter(value)


async def test_get_energy_production_scale_type(energy_production: Node):
    """Test get_energy_production_scale_type function."""
    node = energy_production

    value_id = get_value_id_str(
        node, CommandClass.ENERGY_PRODUCTION, "value", property_key=0
    )
    assert (
        get_energy_production_scale_type(node.values.get(value_id)) == PowerScale.WATTS
    )


async def test_invalid_get_energy_production_scale_type(
    invalid_multilevel_sensor_type: Node,
):
    """Test receiving an invalid energy production scale type."""
    node = invalid_multilevel_sensor_type

    # Create value with an invalid scale ID
    metadata = MetaDataType(
        ccSpecific={CC_SPECIFIC_PARAMETER: 0, CC_SPECIFIC_SCALE: -1}
    )
    value_data = ValueDataType(
        commandClass=CommandClass.ENERGY_PRODUCTION,
        commandClassName="EnergyProduction",
        endpoint=0,
        property="value",
        propertyKey=0,
        metadata=metadata,
    )
    value = Value(node, value_data)
    with pytest.raises(UnknownValueData):
        get_energy_production_scale_type(value)



"""
Constants for the Thermostat CCs.

Includes Thermostat Fan Mode, Thermostat Fan State, Thermostat Mode, Thermostat
Operating State, Thermostat Setback, and Thermostat Setpoint CCs.
"""

from __future__ import annotations

from enum import IntEnum

THERMOSTAT_MODE_PROPERTY = "mode"
THERMOSTAT_SETPOINT_PROPERTY = "setpoint"
THERMOSTAT_OPERATING_STATE_PROPERTY = "state"
THERMOSTAT_CURRENT_TEMP_PROPERTY = "Air temperature"
THERMOSTAT_HUMIDITY_PROPERTY = "Humidity"
THERMOSTAT_FAN_MODE_PROPERTY = "mode"
THERMOSTAT_FAN_OFF_PROPERTY = "off"
THERMOSTAT_FAN_STATE_PROPERTY = "state"


class ThermostatMode(IntEnum):
    """Enum with all (known/used) Z-Wave ThermostatModes."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/ThermostatModeCC.ts#L53-L70
    OFF = 0
    HEAT = 1
    COOL = 2
    AUTO = 3
    AUXILIARY = 4
    RESUME_ON = 5
    FAN = 6
    FURNACE = 7
    DRY = 8
    MOIST = 9
    AUTO_CHANGE_OVER = 10
    HEATING_ECON = 11
    COOLING_ECON = 12
    AWAY = 13
    FULL_POWER = 15
    MANUFACTURER_SPECIFIC = 31


class ThermostatOperatingState(IntEnum):
    """Enum with all (known/used) Z-Wave Thermostat OperatingStates."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/ThermostatOperatingStateCC.ts#L38-L51
    IDLE = 0
    HEATING = 1
    COOLING = 2
    FAN_ONLY = 3
    PENDING_HEAT = 4
    PENDING_COOL = 5
    VENT_ECONOMIZER = 6
    AUX_HEATING = 7
    SECOND_STAGE_HEATING = 8
    SECOND_STAGE_COOLING = 9
    SECOND_STAGE_AUX_HEAT = 10
    THIRD_STAGE_AUX_HEAT = 11


class ThermostatSetpointType(IntEnum):
    """Enum with all (known/used) Z-Wave Thermostat Setpoint Types."""

    # https://github.com/zwave-js/node-zwave-js/blob/master/packages/zwave-js/src/lib/commandclass/ThermostatSetpointCC.ts#L53-L66
    NA = 0
    HEATING = 1
    COOLING = 2
    FURNACE = 7
    DRY_AIR = 8
    MOIST_AIR = 9
    AUTO_CHANGEOVER = 10
    ENERGY_SAVE_HEATING = 11
    ENERGY_SAVE_COOLING = 12
    AWAY_HEATING = 13
    AWAY_COOLING = 14
    FULL_POWER = 15


THERMOSTAT_MODE_SETPOINT_MAP: dict[int, list[ThermostatSetpointType]] = {
    ThermostatMode.OFF: [],
    ThermostatMode.HEAT: [ThermostatSetpointType.HEATING],
    ThermostatMode.COOL: [ThermostatSetpointType.COOLING],
    ThermostatMode.AUTO: [
        ThermostatSetpointType.HEATING,
        ThermostatSetpointType.COOLING,
    ],
    ThermostatMode.AUXILIARY: [ThermostatSetpointType.HEATING],
    ThermostatMode.FURNACE: [ThermostatSetpointType.FURNACE],
    ThermostatMode.DRY: [ThermostatSetpointType.DRY_AIR],
    ThermostatMode.MOIST: [ThermostatSetpointType.MOIST_AIR],
    ThermostatMode.AUTO_CHANGE_OVER: [ThermostatSetpointType.AUTO_CHANGEOVER],
    ThermostatMode.HEATING_ECON: [ThermostatSetpointType.ENERGY_SAVE_HEATING],
    ThermostatMode.COOLING_ECON: [ThermostatSetpointType.ENERGY_SAVE_COOLING],
    ThermostatMode.AWAY: [
        ThermostatSetpointType.AWAY_HEATING,
        ThermostatSetpointType.AWAY_COOLING,
    ],
    ThermostatMode.FULL_POWER: [ThermostatSetpointType.FULL_POWER],
}



"""Test value model."""

from copy import deepcopy

from zwave_js_server.const import ConfigurationValueType, SetValueStatus
from zwave_js_server.model.node import Node
from zwave_js_server.model.value import (
    ConfigurationValue,
    ConfigurationValueFormat,
    MetaDataType,
    SetValueResult,
    ValueDataType,
    get_value_id_str,
)


def test_value_size(lock_schlage_be469):
    """Test the value size property for a value."""
    node = lock_schlage_be469
    zwave_value = node.values["20-112-0-3"]
    assert zwave_value.metadata.value_size == 1


def test_buffer_dict(client, idl_101_lock_state):
    """Test that we handle buffer dictionary correctly."""
    node_data = deepcopy(idl_101_lock_state)
    node = Node(client, node_data)

    value_id = get_value_id_str(node, 99, "userCode", 0, 3)

    assert value_id == "26-99-0-userCode-3"

    zwave_value = node.values[value_id]

    assert zwave_value.metadata.type == "buffer"
    assert zwave_value.value == "¤\x0eªV"


def test_unparseable_value(client, unparseable_json_string_value_state):
    """Test that we handle string value with unparseable format."""
    node = Node(client, unparseable_json_string_value_state)

    value_id = get_value_id_str(node, 99, "userCode", 0, 4)

    assert value_id == "20-99-0-userCode-4"
    assert value_id not in node.values


def test_allow_manual_entry(client, inovelli_switch_state):
    """Test that allow_manaual_entry works correctly."""
    node = Node(client, inovelli_switch_state)

    config_values = node.get_configuration_values()
    value_id = get_value_id_str(node, 112, 8, 0, 255)

    zwave_value = config_values[value_id]

    assert zwave_value.configuration_value_type == ConfigurationValueType.MANUAL_ENTRY

    value_id = get_value_id_str(node, 112, 8, 0, 65280)
    zwave_value = config_values[value_id]

    assert zwave_value.configuration_value_type == ConfigurationValueType.ENUMERATED


def test_stateful(lock_schlage_be469):
    """Test the stateful property for a value."""
    node = lock_schlage_be469
    zwave_value = node.values["20-112-0-3"]
    assert not zwave_value.metadata.secret


def test_secret(lock_schlage_be469):
    """Test the secret property for a value."""
    node = lock_schlage_be469
    zwave_value = node.values["20-112-0-3"]
    assert zwave_value.metadata.stateful


def test_configuration_value_type(inovelli_switch_state):
    """Test configuration value types."""
    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="boolean",
                max=2,
                min=0,
                allowManualEntry=True,
                states={True: "On", False: "Off"},
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.MANUAL_ENTRY

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="boolean",
                max=1,
                min=0,
                allowManualEntry=False,
                states={True: "On", False: "Off"},
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.ENUMERATED

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="boolean",
                max=1,
                min=0,
                allowManualEntry=False,
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.BOOLEAN

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="number",
                max=2,
                min=0,
                allowManualEntry=True,
                states={0: "On", 1: "Off"},
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.MANUAL_ENTRY

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="number",
                max=1,
                min=0,
                allowManualEntry=False,
                states={"1": "On", "0": "Off"},
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.ENUMERATED

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="number",
                max=2,
                min=0,
                allowManualEntry=False,
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.RANGE

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="number",
                allowManualEntry=False,
            ),
        ),
    )
    assert value.configuration_value_type == ConfigurationValueType.UNDEFINED


def test_set_value_result_str():
    """Test SetValueResult str."""
    result = SetValueResult({"status": 255})
    assert result.status == SetValueStatus.SUCCESS
    assert str(result) == "Success"

    result = SetValueResult({"status": 1, "remainingDuration": {"unit": "default"}})
    assert result.status == SetValueStatus.WORKING
    assert str(result) == "Working (default duration)"

    result = SetValueResult(
        {"status": 1, "remainingDuration": {"unit": "seconds", "value": 1}}
    )
    assert result.status == SetValueStatus.WORKING
    assert str(result) == "Working (1 seconds)"

    result = SetValueResult({"status": 3, "message": "test"})
    assert result.status == SetValueStatus.ENDPOINT_NOT_FOUND
    assert str(result) == "Endpoint Not Found: test"

    result = SetValueResult({"status": 1, "remainingDuration": "unknown"})
    assert result.status == SetValueStatus.WORKING
    assert str(result) == "Working (unknown duration)"


def test_configuration_value_metadata(inovelli_switch_state):
    """Test configuration value specific metadata."""
    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="boolean",
                max=2,
                min=0,
                allowManualEntry=True,
                states={True: "On", False: "Off"},
            ),
        ),
    )
    metadata = value.metadata
    assert metadata.default is None
    assert metadata.is_advanced is None
    assert metadata.is_from_config is None
    assert metadata.requires_re_inclusion is None
    assert metadata.no_bulk_support is None
    assert metadata.value_size is None
    assert metadata.format is None

    value = ConfigurationValue(
        inovelli_switch_state,
        ValueDataType(
            commandClass=112,
            property=8,
            propertyName="8",
            endpoint=0,
            metadata=MetaDataType(
                type="boolean",
                max=2,
                min=0,
                allowManualEntry=True,
                states={True: "On", False: "Off"},
                default=2,
                isAdvanced=True,
                isFromConfig=True,
                requiresReInclusion=True,
                noBulkSupport=True,
                valueSize=1,
                format=0,
            ),
        ),
    )
    metadata = value.metadata
    assert metadata.default == 2
    assert metadata.is_advanced
    assert metadata.is_from_config
    assert metadata.requires_re_inclusion
    assert metadata.no_bulk_support
    assert metadata.value_size == 1
    assert metadata.format == ConfigurationValueFormat.SIGNED_INTEGER



"""Helpers for scripts."""

import json
import pathlib
import re
import subprocess
import sys

from const import BRANCH_NAME, GITHUB_PROJECT
from slugify import slugify


def check_dirty_repo():
    """Check if repo is dirty and message user."""
    if subprocess.run(["which", "git"], capture_output=True, check=True).stdout:
        if (
            subprocess.run(
                ["git", "diff", "--stat"],
                check=True,
            ).stdout
            is not None
        ):
            print("Repo is dirty and needs to be committed!")
            sys.exit(1)
    else:
        print(
            "Could not run `git diff --stat` on repo, please run it to determine "
            "whether constants have changed."
        )


def get_registry_location(filename: str) -> str:
    """Get the registry location for the given filename."""
    return (
        f"https://github.com/{GITHUB_PROJECT}/blob/{BRANCH_NAME}/packages/core/"
        f"src/registries/{filename}"
    )


def run_black(file_path: str):
    """Run black on the given file path."""
    if subprocess.run(["which", "black"], capture_output=True, check=True).stdout:
        subprocess.run(
            ["black", file_path],
            check=True,
        )
    else:
        print("Could not run black on new file, please run it to properly format it.")


def get_manually_written_code(file_path: str):
    """Get a list of manually written code from the given file path."""
    existing_const_file = file_path.read_text(encoding="utf-8").splitlines()

    manually_written_code_start_idx = (
        next(
            i
            for i, line in enumerate(existing_const_file)
            if "**END OF AUTOGENERATED CONTENT**" in line
        )
        + 6
    )
    if len(existing_const_file) > manually_written_code_start_idx:
        return [
            line.strip("\n")
            for line in existing_const_file[manually_written_code_start_idx:]
        ]
    return []


def remove_parenthesis(text: str) -> str:
    """Remove text in parenthesis from a string."""
    return re.sub(r"\([^)]*\)", "", text)


def enum_name_format(name: str, should_remove_parenthesis: bool) -> str:
    """Convert sensor/scale name to enum format."""
    if should_remove_parenthesis:
        name = remove_parenthesis(name)
    return slugify(name, separator="_").upper()


def separate_camel_case(my_str: str) -> str:
    """Split a camel case string."""
    if all(char.islower() for char in my_str):
        return my_str.title()
    start_idx = [i for i, e in enumerate(my_str) if e.isupper()] + [len(my_str)]
    start_idx = [0] + start_idx
    return " ".join(my_str[x:y] for x, y in zip(start_idx, start_idx[1:])).title()


def get_json(file_path: str) -> list | dict:
    """Get a JSON dict/list from the given file path."""
    return json.loads(pathlib.Path(file_path).read_text())


def normalize_name(name: str) -> str:
    """Convert a sensor/scale name into a normalized name."""
    return enum_name_format(name, True).replace("_", " ").title()


def format_for_class_name(name: str, suffix: str = "") -> str:
    """Convert sensor/scale name to class name format."""
    return f"{normalize_name(name).replace(' ', '')}{suffix}"



"""Meter Command Class specific utility functions for values."""

from __future__ import annotations

from ...const import CommandClass
from ...const.command_class.meter import (
    CC_SPECIFIC_METER_TYPE,
    CC_SPECIFIC_SCALE,
    METER_TYPE_TO_SCALE_ENUM_MAP,
    MeterScaleType,
    MeterType,
)
from ...exceptions import InvalidCommandClass, UnknownValueData
from ...model.value import Value


def get_meter_type(value: Value) -> MeterType:
    """Get the MeterType for a given value."""
    if value.command_class != CommandClass.METER:
        raise InvalidCommandClass(value, CommandClass.METER)
    try:
        return MeterType(value.metadata.cc_specific[CC_SPECIFIC_METER_TYPE])
    except ValueError:
        raise UnknownValueData(
            value, f"metadata.cc_specific.{CC_SPECIFIC_METER_TYPE}"
        ) from None


def get_meter_scale_type(value: Value) -> MeterScaleType:
    """Get the ScaleType for a given value."""
    meter_type = get_meter_type(value)
    scale_enum = METER_TYPE_TO_SCALE_ENUM_MAP[meter_type]
    try:
        return scale_enum(value.metadata.cc_specific[CC_SPECIFIC_SCALE])
    except ValueError:
        raise UnknownValueData(
            value, f"metadata.cc_specific.{CC_SPECIFIC_SCALE}"
        ) from None



"""Provide Event base classes for Z-Wave JS."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
import logging
from typing import Literal

from pydantic import BaseModel

LOGGER = logging.getLogger(__package__)


class BaseEventModel(BaseModel):
    """Base model for an event."""

    source: Literal["controller", "driver", "node"]
    event: str

    @classmethod
    def from_dict(cls, data: dict) -> BaseEventModel:
        """Initialize from dict."""
        return cls(
            source=data["source"],
            event=data["event"],
        )


@dataclass
class Event:
    """Represent an event."""

    type: str
    data: dict = field(default_factory=dict)


class EventBase:
    """Represent a Z-Wave JS base class for event handling models."""

    def __init__(self) -> None:
        """Initialize event base."""
        self._listeners: dict[str, list[Callable]] = {}

    def on(self, event_name: str, callback: Callable) -> Callable:
        """Register an event callback."""
        listeners: list = self._listeners.setdefault(event_name, [])
        listeners.append(callback)

        def unsubscribe() -> None:
            """Unsubscribe listeners."""
            if callback in listeners:
                listeners.remove(callback)

        return unsubscribe

    def once(self, event_name: str, callback: Callable) -> Callable:
        """Listen for an event exactly once."""

        def event_listener(data: dict) -> None:
            unsub()
            callback(data)

        unsub = self.on(event_name, event_listener)

        return unsub

    def emit(self, event_name: str, data: dict) -> None:
        """Run all callbacks for an event."""
        for listener in self._listeners.get(event_name, []).copy():
            try:
                listener(data)
            except Exception:  # pylint: disable=broad-exception-caught
                LOGGER.exception("Error handling event: %s", event_name)

    def _handle_event_protocol(self, event: Event) -> None:
        """Process an event based on event protocol."""
        handler = getattr(self, f"handle_{event.type.replace(' ', '_')}", None)
        if handler is None:
            LOGGER.debug("Received unknown event: %s", event)
            return
        handler(event)  # pylint: disable=not-callable



"""Provide a model for the Z-Wave JS controller."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, cast

from zwave_js_server.model.node.firmware import NodeFirmwareUpdateInfo

from ...const import (
    MINIMUM_QR_STRING_LENGTH,
    AssociationCheckResult,
    ControllerStatus,
    ExclusionStrategy,
    InclusionState,
    InclusionStrategy,
    NodeType,
    QRCodeVersion,
    RFRegion,
    RemoveNodeReason,
    ZwaveFeature,
)
from ...event import Event, EventBase
from ...util.helpers import convert_base64_to_bytes, convert_bytes_to_base64
from ..association import AssociationAddress, AssociationGroup
from ..node import Node
from ..node.firmware import NodeFirmwareUpdateResult
from .data_model import ControllerDataType
from .event_model import CONTROLLER_EVENT_MODEL_MAP
from .inclusion_and_provisioning import (
    InclusionGrant,
    ProvisioningEntry,
    QRProvisioningInformation,
)
from .rebuild_routes import (
    RebuildRoutesOptions,
    RebuildRoutesOptionsDataType,
    RebuildRoutesStatus,
)
from .statistics import (
    ControllerLifelineRoutes,
    ControllerStatistics,
    ControllerStatisticsDataType,
)

if TYPE_CHECKING:
    from ...client import Client


DEFAULT_CONTROLLER_STATISTICS = ControllerStatisticsDataType(
    messagesTX=0,
    messagesRX=0,
    messagesDroppedTX=0,
    messagesDroppedRX=0,
    NAK=0,
    CAN=0,
    timeoutACK=0,
    timeoutResponse=0,
    timeoutCallback=0,
)


@dataclass
class NVMProgress:
    """Class to represent an NVM backup/restore progress event."""

    bytes_read_or_written: int
    total_bytes: int


class Controller(EventBase):
    """Represent a Z-Wave JS controller."""

    def __init__(self, client: Client, state: dict) -> None:
        """Initialize controller."""
        super().__init__()
        self.client = client
        self.nodes: dict[int, Node] = {}
        self._rebuild_routes_progress: dict[Node, RebuildRoutesStatus] | None = None
        self._last_rebuild_routes_result: dict[Node, RebuildRoutesStatus] | None = None
        self._statistics = ControllerStatistics(DEFAULT_CONTROLLER_STATISTICS)
        for node_state in state["nodes"]:
            node = Node(client, node_state)
            self.nodes[node.node_id] = node
        self.update(state["controller"])

    def __repr__(self) -> str:
        """Return the representation."""
        return f"{type(self).__name__}(home_id={self.home_id})"

    def __hash__(self) -> int:
        """Return the hash."""
        return hash(self.home_id)

    def __eq__(self, other: object) -> bool:
        """Return whether this instance equals another."""
        if not isinstance(other, Controller):
            return False
        return self.home_id == other.home_id

    def _generate_rebuild_routes_status(
        self, data: dict[str, str]
    ) -> dict[Node, RebuildRoutesStatus]:
        """Generate rebuild routes status."""
        return {
            self.nodes[int(node_id)]: RebuildRoutesStatus(status)
            for node_id, status in data.items()
        }

    @property
    def sdk_version(self) -> str | None:
        """Return sdk_version."""
        return self.data.get("sdkVersion")

    @property
    def controller_type(self) -> int | None:
        """Return controller_type."""
        return self.data.get("type")

    @property
    def home_id(self) -> int | None:
        """Return home_id."""
        return self.data.get("homeId")

    @property
    def own_node_id(self) -> int | None:
        """Return own_node_id."""
        return self.data.get("ownNodeId")

    @property
    def own_node(self) -> Node | None:
        """Return own_node."""
        if self.own_node_id is None:
            return None
        return self.nodes.get(self.own_node_id)

    @property
    def is_primary(self) -> bool | None:
        """Return is_primary."""
        return self.data.get("isPrimary")

    @property
    def is_using_home_id_from_other_network(self) -> bool | None:
        """Return is_using_home_id_from_other_network."""
        return self.data.get("isUsingHomeIdFromOtherNetwork")

    @property
    def is_SIS_present(self) -> bool | None:  # pylint: disable=invalid-name
        """Return is_SIS_present."""
        return self.data.get("isSISPresent")

    @property
    def was_real_primary(self) -> bool | None:
        """Return was_real_primary."""
        return self.data.get("wasRealPrimary")

    @property
    def is_suc(self) -> bool | None:
        """Return is_suc."""
        return self.data.get("isSUC")

    @property
    def node_type(self) -> NodeType | None:
        """Return node_type."""
        if (node_type := self.data.get("nodeType")) is not None:
            return NodeType(node_type)
        return None

    @property
    def firmware_version(self) -> str | None:
        """Return firmware_version."""
        return self.data.get("firmwareVersion")

    @property
    def manufacturer_id(self) -> int | None:
        """Return manufacturer_id."""
        return self.data.get("manufacturerId")

    @property
    def product_type(self) -> int | None:
        """Return product_type."""
        return self.data.get("productType")

    @property
    def product_id(self) -> int | None:
        """Return product_id."""
        return self.data.get("productId")

    @property
    def supported_function_types(self) -> list[int]:
        """Return supported_function_types."""
        return self.data.get("supportedFunctionTypes", [])

    @property
    def suc_node_id(self) -> int | None:
        """Return suc_node_id."""
        return self.data.get("sucNodeId")

    @property
    def supports_timers(self) -> bool | None:
        """Return supports_timers."""
        return self.data.get("supportsTimers")

    @property
    def is_rebuilding_routes(self) -> bool | None:
        """Return is_rebuilding_routes."""
        return self.data.get("isRebuildingRoutes")

    @property
    def statistics(self) -> ControllerStatistics:
        """Return statistics property."""
        return self._statistics

    @property
    def rebuild_routes_progress(self) -> dict[Node, RebuildRoutesStatus] | None:
        """Return rebuild routes progress state."""
        return self._rebuild_routes_progress

    @property
    def last_rebuild_routes_result(self) -> dict[Node, RebuildRoutesStatus] | None:
        """Return the last rebuild routes result."""
        return self._last_rebuild_routes_result

    @property
    def inclusion_state(self) -> InclusionState:
        """Return inclusion state."""
        return InclusionState(self.data["inclusionState"])

    @property
    def rf_region(self) -> RFRegion | None:
        """Return RF region of controller."""
        if (rf_region := self.data.get("rfRegion")) is None:
            return None
        return RFRegion(rf_region)

    @property
    def status(self) -> ControllerStatus:
        """Return status."""
        return ControllerStatus(self.data["status"])

    @property
    def supports_long_range(self) -> bool | None:
        """Return whether controller supports long range or not."""
        return self.data.get("supportsLongRange")

    def update(self, data: ControllerDataType) -> None:
        """Update controller data."""
        self.data = data
        self._statistics = ControllerStatistics(
            self.data.get("statistics", DEFAULT_CONTROLLER_STATISTICS)
        )
        if "rebuildRoutesProgress" in self.data:
            self._rebuild_routes_progress = self._generate_rebuild_routes_status(
                self.data["rebuildRoutesProgress"]
            )

    async def async_begin_inclusion(
        self,
        inclusion_strategy: Literal[
            InclusionStrategy.DEFAULT,
            InclusionStrategy.SECURITY_S0,
            InclusionStrategy.SECURITY_S2,
            InclusionStrategy.INSECURE,
        ],
        force_security: bool | None = None,
        provisioning: str | ProvisioningEntry | QRProvisioningInformation | None = None,
        dsk: str | None = None,
    ) -> bool:
        """Send beginInclusion command to Controller."""
        # Most functionality was introduced in Schema 8
        require_schema = 8
        options: dict[str, Any] = {"strategy": inclusion_strategy}
        # forceSecurity can only be used with the default inclusion strategy
        if force_security is not None:
            if inclusion_strategy != InclusionStrategy.DEFAULT:
                raise ValueError(
                    "`forceSecurity` option is only supported with inclusion_strategy="
                    "DEFAULT"
                )
            options["forceSecurity"] = force_security

        # provisioning can only be used with the S2 inclusion strategy and may need
        # additional processing
        if provisioning is not None:
            if inclusion_strategy != InclusionStrategy.SECURITY_S2:
                raise ValueError(
                    "`provisioning` option is only supported with inclusion_strategy="
                    "SECURITY_S2"
                )

            if dsk is not None:
                raise ValueError("Only one of `provisioning` and `dsk` can be provided")
            # Provisioning option was introduced in Schema 11
            require_schema = 11
            # String is assumed to be the QR code string so we can pass as is
            if isinstance(provisioning, str):
                if len(
                    provisioning
                ) < MINIMUM_QR_STRING_LENGTH or not provisioning.startswith("90"):
                    raise ValueError(
                        f"QR code string must be at least {MINIMUM_QR_STRING_LENGTH} "
                        "characters long and start with `90`"
                    )
                options["provisioning"] = provisioning
            # If we get a Smart Start QR code, we provision the node and return because
            # inclusion is over
            elif (
                isinstance(provisioning, QRProvisioningInformation)
                and provisioning.version == QRCodeVersion.SMART_START
            ):
                raise ValueError(
                    "Smart Start QR codes can't use the normal inclusion process. Use "
                    "the provision_smart_start_node command to provision this device."
                )
            # Otherwise we assume the data is ProvisioningEntry or
            # QRProvisioningInformation that is not a Smart Start QR code
            else:
                options["provisioning"] = provisioning.to_dict()

        if dsk is not None:
            if inclusion_strategy != InclusionStrategy.SECURITY_S2:
                raise ValueError(
                    "`dsk` option is only supported with inclusion_strategy=SECURITY_S2"
                )

            require_schema = 25
            options["dsk"] = dsk

        data = await self.client.async_send_command(
            {
                "command": "controller.begin_inclusion",
                "options": options,
            },
            require_schema=require_schema,
        )
        return cast(bool, data["success"])

    async def async_provision_smart_start_node(
        self,
        provisioning_info: ProvisioningEntry | QRProvisioningInformation | str,
    ) -> None:
        """Send provisionSmartStartNode command to Controller."""
        if (
            isinstance(provisioning_info, QRProvisioningInformation)
            and provisioning_info.version == QRCodeVersion.S2
        ):
            raise ValueError(
                "An S2 QR Code can't be used to pre-provision a Smart Start node"
            )
        await self.client.async_send_command(
            {
                "command": "controller.provision_smart_start_node",
                "entry": (
                    provisioning_info
                    if isinstance(provisioning_info, str)
                    else provisioning_info.to_dict()
                ),
            },
            require_schema=11,
        )

    async def async_unprovision_smart_start_node(
        self, dsk_or_node_id: int | str
    ) -> None:
        """Send unprovisionSmartStartNode command to Controller."""
        await self.client.async_send_command(
            {
                "command": "controller.unprovision_smart_start_node",
                "dskOrNodeId": dsk_or_node_id,
            },
            require_schema=11,
        )

    async def async_get_provisioning_entry(
        self, dsk_or_node_id: int | str
    ) -> ProvisioningEntry | None:
        """Send getProvisioningEntry command to Controller."""
        data = await self.client.async_send_command(
            {
                "command": "controller.get_provisioning_entry",
                "dskOrNodeId": dsk_or_node_id,
            },
            require_schema=17,
        )
        if "entry" in data:
            return ProvisioningEntry.from_dict(data["entry"])
        return None

    async def async_get_provisioning_entries(self) -> list[ProvisioningEntry]:
        """Send getProvisioningEntries command to Controller."""
        data = await self.client.async_send_command(
            {
                "command": "controller.get_provisioning_entries",
            },
            require_schema=11,
        )
        return [ProvisioningEntry.from_dict(entry) for entry in data.get("entries", [])]

    async def async_stop_inclusion(self) -> bool:
        """Send stopInclusion command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.stop_inclusion"}
        )
        return cast(bool, data["success"])

    async def async_cancel_secure_bootstrap_s2(self) -> None:
        """Send cancelSecureBootstrapS2 command to Controller."""
        await self.client.async_send_command(
            {"command": "controller.cancel_secure_bootstrap_s2"}, require_schema=40
        )

    async def async_begin_exclusion(
        self, strategy: ExclusionStrategy | None = None
    ) -> bool:
        """Send beginExclusion command to Controller."""
        payload: dict[str, str | dict[str, ExclusionStrategy]] = {
            "command": "controller.begin_exclusion"
        }
        if strategy is not None:
            payload["options"] = {"strategy": strategy}
        data = await self.client.async_send_command(payload, require_schema=22)
        return cast(bool, data["success"])

    async def async_stop_exclusion(self) -> bool:
        """Send stopExclusion command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.stop_exclusion"}
        )
        return cast(bool, data["success"])

    async def async_remove_failed_node(self, node: Node) -> None:
        """Send removeFailedNode command to Controller."""
        await self.client.async_send_command(
            {"command": "controller.remove_failed_node", "nodeId": node.node_id}
        )

    async def async_replace_failed_node(
        self,
        node: Node,
        inclusion_strategy: Literal[
            InclusionStrategy.DEFAULT,
            InclusionStrategy.SECURITY_S0,
            InclusionStrategy.SECURITY_S2,
            InclusionStrategy.INSECURE,
        ],
        force_security: bool | None = None,
        provisioning: str | ProvisioningEntry | QRProvisioningInformation | None = None,
    ) -> bool:
        """Send replaceFailedNode command to Controller."""
        # Most functionality was introduced in Schema 8
        require_schema = 8
        options: dict[str, Any] = {"strategy": inclusion_strategy}
        # forceSecurity can only be used with the default inclusion strategy
        if force_security is not None:
            if inclusion_strategy != InclusionStrategy.DEFAULT:
                raise ValueError(
                    "`forceSecurity` option is only supported with inclusion_strategy="
                    "DEFAULT"
                )
            options["forceSecurity"] = force_security

        # provisioning can only be used with the S2 inclusion strategy and may need
        # additional processing
        if provisioning is not None:
            if inclusion_strategy != InclusionStrategy.SECURITY_S2:
                raise ValueError(
                    "`provisioning` option is only supported with inclusion_strategy="
                    "SECURITY_S2"
                )
            # Provisioning option was introduced in Schema 11
            require_schema = 11
            # String is assumed to be the QR code string so we can pass as is
            if isinstance(provisioning, str):
                if len(
                    provisioning
                ) < MINIMUM_QR_STRING_LENGTH or not provisioning.startswith("90"):
                    raise ValueError(
                        f"QR code string must be at least {MINIMUM_QR_STRING_LENGTH} "
                        "characters long and start with `90`"
                    )
                options["provisioning"] = provisioning
            # Otherwise we assume the data is ProvisioningEntry or
            # QRProvisioningInformation
            else:
                options["provisioning"] = provisioning.to_dict()

        data = await self.client.async_send_command(
            {
                "command": "controller.replace_failed_node",
                "nodeId": node.node_id,
                "options": options,
            },
            require_schema=require_schema,
        )
        return cast(bool, data["success"])

    async def async_rebuild_node_routes(self, node: Node) -> bool:
        """Send rebuildNodeRoutes command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.rebuild_node_routes", "nodeId": node.node_id},
            require_schema=32,
        )
        return cast(bool, data["success"])

    async def async_begin_rebuilding_routes(
        self, options: RebuildRoutesOptions | None = None
    ) -> bool:
        """Send beginRebuildingRoutes command to Controller."""
        msg: dict[str, str | RebuildRoutesOptionsDataType] = {
            "command": "controller.begin_rebuilding_routes"
        }
        if options:
            msg["options"] = options.to_dict()
        data = await self.client.async_send_command(msg, require_schema=32)
        return cast(bool, data["success"])

    async def async_stop_rebuilding_routes(self) -> bool:
        """Send stopRebuildingRoutes command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.stop_rebuilding_routes"}, require_schema=32
        )
        success = cast(bool, data["success"])
        if success:
            self._rebuild_routes_progress = None
            self.data["isRebuildingRoutes"] = False
        return success

    async def async_is_failed_node(self, node: Node) -> bool:
        """Send isFailedNode command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.is_failed_node", "nodeId": node.node_id}
        )
        return cast(bool, data["failed"])

    async def async_get_association_groups(
        self, source: AssociationAddress
    ) -> dict[int, AssociationGroup]:
        """Send getAssociationGroups command to Controller."""
        source_data = {"nodeId": source.node_id}
        if source.endpoint is not None:
            source_data["endpoint"] = source.endpoint
        data = await self.client.async_send_command(
            {
                "command": "controller.get_association_groups",
                **source_data,
            }
        )
        groups = {}
        for key, group in data["groups"].items():
            groups[int(key)] = AssociationGroup(
                max_nodes=group["maxNodes"],
                is_lifeline=group["isLifeline"],
                multi_channel=group["multiChannel"],
                label=group["label"],
                profile=group.get("profile"),
                issued_commands=group.get("issuedCommands", {}),
            )
        return groups

    async def async_get_associations(
        self, source: AssociationAddress
    ) -> dict[int, list[AssociationAddress]]:
        """Send getAssociations command to Controller."""
        source_data = {"nodeId": source.node_id}
        if source.endpoint is not None:
            source_data["endpoint"] = source.endpoint
        data = await self.client.async_send_command(
            {
                "command": "controller.get_associations",
                **source_data,
            }
        )
        associations = {}
        for key, association_addresses in data["associations"].items():
            associations[int(key)] = [
                AssociationAddress(
                    self,
                    node_id=association_address["nodeId"],
                    endpoint=association_address.get("endpoint"),
                )
                for association_address in association_addresses
            ]
        return associations

    async def async_check_association(
        self, source: AssociationAddress, group: int, association: AssociationAddress
    ) -> AssociationCheckResult:
        """Send checkAssociation command to Controller."""
        source_data = {"nodeId": source.node_id}
        if source.endpoint is not None:
            source_data["endpoint"] = source.endpoint

        association_data = {"nodeId": association.node_id}
        if association.endpoint is not None:
            association_data["endpoint"] = association.endpoint
        data = await self.client.async_send_command(
            {
                "command": "controller.check_association",
                **source_data,
                "group": group,
                "association": association_data,
            },
            require_schema=37,
        )
        return AssociationCheckResult(data["result"])

    async def async_add_associations(
        self,
        source: AssociationAddress,
        group: int,
        associations: list[AssociationAddress],
        wait_for_result: bool = False,
    ) -> None:
        """Send addAssociations command to Controller."""
        source_data = {"nodeId": source.node_id}
        if source.endpoint is not None:
            source_data["endpoint"] = source.endpoint

        associations_data = []
        for association in associations:
            association_data = {"nodeId": association.node_id}
            if association.endpoint is not None:
                association_data["endpoint"] = association.endpoint
            associations_data.append(association_data)

        cmd = {
            "command": "controller.add_associations",
            **source_data,
            "group": group,
            "associations": associations_data,
        }
        if wait_for_result:
            await self.client.async_send_command(cmd)
        else:
            await self.client.async_send_command_no_wait(cmd)

    async def async_remove_associations(
        self,
        source: AssociationAddress,
        group: int,
        associations: list[AssociationAddress],
        wait_for_result: bool = False,
    ) -> None:
        """Send removeAssociations command to Controller."""
        source_data = {"nodeId": source.node_id}
        if source.endpoint is not None:
            source_data["endpoint"] = source.endpoint

        associations_data = []
        for association in associations:
            association_data = {"nodeId": association.node_id}
            if association.endpoint is not None:
                association_data["endpoint"] = association.endpoint
            associations_data.append(association_data)

        cmd = {
            "command": "controller.remove_associations",
            **source_data,
            "group": group,
            "associations": associations_data,
        }
        if wait_for_result:
            await self.client.async_send_command(cmd)
        else:
            await self.client.async_send_command_no_wait(cmd)

    async def async_remove_node_from_all_associations(
        self,
        node: Node,
        wait_for_result: bool = False,
    ) -> None:
        """Send removeNodeFromAllAssociations command to Controller."""
        cmd = {
            "command": "controller.remove_node_from_all_associations",
            "nodeId": node.node_id,
        }
        if wait_for_result:
            await self.client.async_send_command(cmd)
        else:
            await self.client.async_send_command_no_wait(cmd)

    async def async_get_node_neighbors(self, node: Node) -> list[int]:
        """Send getNodeNeighbors command to Controller to get node's neighbors."""
        data = await self.client.async_send_command(
            {
                "command": "controller.get_node_neighbors",
                "nodeId": node.node_id,
            }
        )
        return cast(list[int], data["neighbors"])

    async def async_grant_security_classes(
        self, inclusion_grant: InclusionGrant
    ) -> None:
        """Send grantSecurityClasses command to Controller."""
        await self.client.async_send_command(
            {
                "command": "controller.grant_security_classes",
                "inclusionGrant": inclusion_grant.to_dict(),
            }
        )

    async def async_validate_dsk_and_enter_pin(self, pin: str) -> None:
        """Send validateDSKAndEnterPIN command to Controller."""
        await self.client.async_send_command(
            {
                "command": "controller.validate_dsk_and_enter_pin",
                "pin": pin,
            }
        )

    async def async_supports_feature(self, feature: ZwaveFeature) -> bool | None:
        """
        Send supportsFeature command to Controller.

        When None is returned it means the driver does not yet know whether the
        controller supports the input feature.
        """
        data = await self.client.async_send_command(
            {"command": "controller.supports_feature", "feature": feature.value},
            require_schema=12,
        )
        return cast(bool | None, data.get("supported"))

    async def async_get_state(self) -> ControllerDataType:
        """Get controller state."""
        data = await self.client.async_send_command(
            {"command": "controller.get_state"}, require_schema=14
        )
        return cast(ControllerDataType, data["state"])

    async def async_backup_nvm_raw(self) -> bytes:
        """Send backupNVMRaw command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.backup_nvm_raw"}, require_schema=14
        )
        return convert_base64_to_bytes(data["nvmData"])

    async def async_restore_nvm(
        self, file: bytes, options: dict[str, bool] | None = None
    ) -> None:
        """Send restoreNVM command to Controller."""
        await self.client.async_send_command(
            {
                "command": "controller.restore_nvm",
                "nvmData": convert_bytes_to_base64(file),
                "migrateOptions": {} if options is None else options,
            },
            require_schema=42,
        )

    async def async_backup_nvm_raw_base64(self) -> str:
        """Send backupNVMRaw command to Controller and return base64 string directly."""
        data = await self.client.async_send_command(
            {"command": "controller.backup_nvm_raw"}, require_schema=14
        )
        return data["nvmData"]

    async def async_restore_nvm_base64(
        self, base64_data: str, options: dict[str, bool] | None = None
    ) -> None:
        """Send restoreNVM command to Controller with base64 data directly."""
        await self.client.async_send_command(
            {
                "command": "controller.restore_nvm",
                "nvmData": base64_data,
                "migrateOptions": {} if options is None else options,
            },
            require_schema=42,
        )

    async def async_get_power_level(self) -> dict[str, int]:
        """Send getPowerlevel command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.get_powerlevel"}, require_schema=14
        )
        return {
            "power_level": data["powerlevel"],
            "measured_0_dbm": data["measured0dBm"],
        }

    async def async_set_power_level(
        self, power_level: int, measured_0_dbm: int
    ) -> bool:
        """Send setPowerlevel command to Controller."""
        data = await self.client.async_send_command(
            {
                "command": "controller.set_powerlevel",
                "powerlevel": power_level,
                "measured0dBm": measured_0_dbm,
            },
            require_schema=14,
        )
        return cast(bool, data["success"])

    async def async_get_rf_region(self) -> RFRegion:
        """Send getRFRegion command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.get_rf_region"}, require_schema=14
        )
        return RFRegion(data["region"])

    async def async_set_rf_region(self, rf_region: RFRegion) -> bool:
        """Send setRFRegion command to Controller."""
        data = await self.client.async_send_command(
            {
                "command": "controller.set_rf_region",
                "region": rf_region.value,
            },
            require_schema=14,
        )
        return cast(bool, data["success"])

    async def async_toggle_rf(self, enable: bool) -> bool:
        """Send toggleRF command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.toggle_rf", "enable": enable}, require_schema=43
        )
        return cast(bool, data["success"])

    async def async_get_known_lifeline_routes(
        self,
    ) -> dict[Node, ControllerLifelineRoutes]:
        """Send getKnownLifelineRoutes command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.get_known_lifeline_routes"}, require_schema=16
        )

        return {
            self.nodes[int(node_id)]: ControllerLifelineRoutes(
                self.client, lifeline_routes
            )
            for node_id, lifeline_routes in data["routes"].items()
        }

    async def async_is_any_ota_firmware_update_in_progress(self) -> bool:
        """
        Send isAnyOTAFirmwareUpdateInProgress command to Controller.

        If `True`, a firmware update is in progress on at least one node.
        """
        data = await self.client.async_send_command(
            {"command": "controller.is_any_ota_firmware_update_in_progress"},
            require_schema=21,
        )
        assert data
        return cast(bool, data["progress"])

    async def async_get_available_firmware_updates(
        self, node: Node, api_key: str, include_prereleases: bool = True
    ) -> list[NodeFirmwareUpdateInfo]:
        """Send getAvailableFirmwareUpdates command to Controller."""
        data = await self.client.async_send_command(
            {
                "command": "controller.get_available_firmware_updates",
                "nodeId": node.node_id,
                "apiKey": api_key,
                "includePrereleases": include_prereleases,
            },
            require_schema=32,
        )
        assert data
        return [NodeFirmwareUpdateInfo.from_dict(update) for update in data["updates"]]

    async def async_firmware_update_ota(
        self, node: Node, update_info: NodeFirmwareUpdateInfo
    ) -> NodeFirmwareUpdateResult:
        """Send firmwareUpdateOTA command to Controller."""
        data = await self.client.async_send_command(
            {
                "command": "controller.firmware_update_ota",
                "nodeId": node.node_id,
                "updateInfo": update_info.to_dict(),
            },
            require_schema=32,
        )
        return NodeFirmwareUpdateResult(node, data["result"])

    async def async_is_firmware_update_in_progress(self) -> bool:
        """Send isFirmwareUpdateInProgress command to Controller."""
        data = await self.client.async_send_command(
            {"command": "controller.is_firmware_update_in_progress"}, require_schema=26
        )
        return cast(bool, data["progress"])

    def receive_event(self, event: Event) -> None:
        """Receive an event."""
        if event.data["source"] == "node":
            node = self.nodes.get(event.data["nodeId"])
            if node is None:
                # TODO handle event for unknown node
                pass
            else:
                node.receive_event(event)
            return

        if event.data["source"] != "controller":
            # TODO decide what to do here
            print(
                "Controller doesn't know how to handle/forward this event: "
                f"{event.data}"
            )

        CONTROLLER_EVENT_MODEL_MAP[event.type].from_dict(event.data)

        self._handle_event_protocol(event)

        event.data["controller"] = self
        self.emit(event.type, event.data)

    def handle_inclusion_failed(self, event: Event) -> None:
        """Process an inclusion failed event."""

    def handle_exclusion_failed(self, event: Event) -> None:
        """Process an exclusion failed event."""

    def handle_inclusion_started(self, event: Event) -> None:
        """Process an inclusion started event."""

    def handle_inclusion_state_changed(self, event: Event) -> None:
        """Process an inclusion state changed event."""
        self.data["inclusionState"] = event.data["state"]

    def handle_exclusion_started(self, event: Event) -> None:
        """Process an exclusion started event."""

    def handle_inclusion_stopped(self, event: Event) -> None:
        """Process an inclusion stopped event."""

    def handle_exclusion_stopped(self, event: Event) -> None:
        """Process an exclusion stopped event."""

    def handle_node_found(self, event: Event) -> None:
        """Process a node found event."""

    def handle_node_added(self, event: Event) -> None:
        """Process a node added event."""
        node = event.data["node"] = Node(self.client, event.data["node"])
        self.nodes[node.node_id] = node

    def handle_node_removed(self, event: Event) -> None:
        """Process a node removed event."""
        event.data["reason"] = RemoveNodeReason(event.data["reason"])
        event.data["node"] = self.nodes.pop(event.data["node"]["nodeId"])
        # Remove client from node since it's no longer connected to the controller
        event.data["node"].client = None

    def handle_rebuild_routes_progress(self, event: Event) -> None:
        """Process a rebuild routes progress event."""
        self._rebuild_routes_progress = self._generate_rebuild_routes_status(
            event.data["progress"]
        )
        self.data["isRebuildingRoutes"] = True

    def handle_rebuild_routes_done(self, event: Event) -> None:
        """Process a rebuild routes done event."""
        self._last_rebuild_routes_result = self._generate_rebuild_routes_status(
            event.data["result"]
        )
        self._rebuild_routes_progress = None
        self.data["isRebuildingRoutes"] = False

    def handle_statistics_updated(self, event: Event) -> None:
        """Process a statistics updated event."""
        self.data["statistics"] = statistics = event.data["statistics"]
        self._statistics = event.data["statistics_updated"] = ControllerStatistics(
            statistics
        )

    def handle_grant_security_classes(self, event: Event) -> None:
        """Process a grant security classes event."""
        event.data["requested_grant"] = InclusionGrant.from_dict(
            event.data["requested"]
        )

    def handle_validate_dsk_and_enter_pin(self, event: Event) -> None:
        """Process a validate dsk and enter pin event."""

    def handle_inclusion_aborted(self, event: Event) -> None:
        """Process an inclusion aborted event."""

    def handle_nvm_backup_progress(self, event: Event) -> None:
        """Process a nvm backup progress event."""
        event.data["nvm_backup_progress"] = NVMProgress(
            event.data["bytesRead"], event.data["total"]
        )

    def handle_nvm_convert_progress(self, event: Event) -> None:
        """Process a nvm convert progress event."""
        event.data["nvm_convert_progress"] = NVMProgress(
            event.data["bytesRead"], event.data["total"]
        )

    def handle_nvm_restore_progress(self, event: Event) -> None:
        """Process a nvm restore progress event."""
        event.data["nvm_restore_progress"] = NVMProgress(
            event.data["bytesWritten"], event.data["total"]
        )

    def handle_identify(self, event: Event) -> None:
        """Process an identify event."""
        # TODO handle event for unknown node
        if node := self.nodes.get(event.data["nodeId"]):
            event.data["node"] = node

    def handle_status_changed(self, event: Event) -> None:
        """Process a status changed event."""
        self.data["status"] = event.data["status"]
        event.data["status"] = ControllerStatus(event.data["status"])



"""Provide a model for the Z-Wave JS node."""

from __future__ import annotations

import asyncio
import copy
from datetime import datetime
import logging
from typing import TYPE_CHECKING, Any, Literal, cast

from ...const import (
    INTERVIEW_FAILED,
    NOT_INTERVIEWED,
    CommandClass,
    DateAndTime,
    NodeStatus,
    PowerLevel,
    Protocols,
    SecurityClass,
)
from ...event import Event, EventBase
from ...exceptions import NotFoundError, UnparseableValue, UnwriteableValue
from ..command_class import CommandClassInfo
from ..device_class import DeviceClass
from ..device_config import DeviceConfig
from ..endpoint import Endpoint, EndpointDataType
from ..notification import (
    EntryControlNotification,
    EntryControlNotificationDataType,
    MultilevelSwitchNotification,
    MultilevelSwitchNotificationDataType,
    NotificationNotification,
    NotificationNotificationDataType,
    PowerLevelNotification,
    PowerLevelNotificationDataType,
)
from ..value import (
    ConfigurationValue,
    ConfigurationValueFormat,
    MetaDataType,
    SetConfigParameterResult,
    SetValueResult,
    Value,
    ValueDataType,
    ValueMetadata,
    ValueNotification,
    _get_value_id_str_from_dict,
)
from .data_model import NodeDataType
from .event_model import NODE_EVENT_MODEL_MAP
from .firmware import (
    NodeFirmwareUpdateCapabilities,
    NodeFirmwareUpdateCapabilitiesDataType,
    NodeFirmwareUpdateProgress,
    NodeFirmwareUpdateProgressDataType,
    NodeFirmwareUpdateResult,
    NodeFirmwareUpdateResultDataType,
)
from .health_check import (
    CheckHealthProgress,
    LifelineHealthCheckResult,
    LifelineHealthCheckSummary,
    RouteHealthCheckResult,
    RouteHealthCheckSummary,
    TestPowerLevelProgress,
)
from .statistics import NodeStatistics, NodeStatisticsDataType

if TYPE_CHECKING:
    from ...client import Client


# pylint: disable=too-many-lines
_LOGGER = logging.getLogger(__package__)

DEFAULT_NODE_STATISTICS = NodeStatisticsDataType(
    commandsTX=0,
    commandsRX=0,
    commandsDroppedTX=0,
    commandsDroppedRX=0,
    timeoutResponse=0,
)


def _get_value_id_dict_from_value_data(value_data: ValueDataType) -> dict[str, Any]:
    """Return a value ID dict from ValueDataType."""
    data = {
        "commandClass": value_data["commandClass"],
        "property": value_data["property"],
    }

    if (endpoint := value_data.get("endpoint")) is not None:
        data["endpoint"] = endpoint
    if (property_key := value_data.get("propertyKey")) is not None:
        data["propertyKey"] = property_key

    return data


class Node(EventBase):
    """Represent a Z-Wave JS node."""

    def __init__(self, client: Client, data: NodeDataType) -> None:
        """Initialize the node."""
        super().__init__()
        self.client = client
        self.data: NodeDataType = {}
        self._device_config = DeviceConfig({})
        self._statistics = NodeStatistics(
            client, data.get("statistics", DEFAULT_NODE_STATISTICS)
        )
        self._firmware_update_progress: NodeFirmwareUpdateProgress | None = None
        self._device_class: DeviceClass | None = None
        self._last_seen: datetime | None = None
        self.values: dict[str, ConfigurationValue | Value] = {}
        self.endpoints: dict[int, Endpoint] = {}
        self.status_event = asyncio.Event()
        self.update(data)

    def __repr__(self) -> str:
        """Return the representation."""
        return f"{type(self).__name__}(node_id={self.node_id})"

    def __hash__(self) -> int:
        """Return the hash."""
        return hash((self.client.driver, self.node_id))

    def __eq__(self, other: object) -> bool:
        """Return whether this instance equals another."""
        if not isinstance(other, Node):
            return False
        return (
            self.client.driver == other.client.driver and self.node_id == other.node_id
        )

    def _init_value(self, val: ValueDataType) -> Value | ConfigurationValue:
        """Initialize a Value object from ValueDataType."""
        if val["commandClass"] == CommandClass.CONFIGURATION:
            return ConfigurationValue(self, val)
        return Value(self, val)

    @property
    def node_id(self) -> int:
        """Return node ID property."""
        return self.data["nodeId"]

    @property
    def index(self) -> int:
        """Return index property."""
        return self.data["index"]

    @property
    def device_class(self) -> DeviceClass | None:
        """Return the device_class."""
        return self._device_class

    @property
    def installer_icon(self) -> int | None:
        """Return installer icon property."""
        return self.data.get("installerIcon")

    @property
    def user_icon(self) -> int | None:
        """Return user icon property."""
        return self.data.get("userIcon")

    @property
    def status(self) -> NodeStatus:
        """Return the status."""
        return NodeStatus(self.data["status"])

    @property
    def ready(self) -> bool | None:
        """Return the ready."""
        return self.data.get("ready")

    @property
    def is_listening(self) -> bool | None:
        """Return the is_listening."""
        return self.data.get("isListening")

    @property
    def is_frequent_listening(self) -> bool | str | None:
        """Return the is_frequent_listening."""
        return self.data.get("isFrequentListening")

    @property
    def is_routing(self) -> bool | None:
        """Return the is_routing."""
        return self.data.get("isRouting")

    @property
    def max_data_rate(self) -> int | None:
        """Return the max_data_rate."""
        return self.data.get("maxDataRate")

    @property
    def supported_data_rates(self) -> list[int]:
        """Return the supported_data_rates."""
        return self.data.get("supportedDataRates", [])

    @property
    def is_secure(self) -> bool | None:
        """Return the is_secure."""
        if (is_secure := self.data.get("isSecure")) is not None:
            return is_secure
        return None

    @property
    def protocol_version(self) -> int | None:
        """Return the protocol_version."""
        return self.data.get("protocolVersion")

    @property
    def supports_beaming(self) -> bool | None:
        """Return the supports_beaming."""
        return self.data.get("supportsBeaming")

    @property
    def supports_security(self) -> bool | None:
        """Return the supports_security."""
        return self.data.get("supportsSecurity")

    @property
    def manufacturer_id(self) -> int | None:
        """Return the manufacturer_id."""
        return self.data.get("manufacturerId")

    @property
    def product_id(self) -> int | None:
        """Return the product_id."""
        return self.data.get("productId")

    @property
    def product_type(self) -> int | None:
        """Return the product_type."""
        return self.data.get("productType")

    @property
    def firmware_version(self) -> str | None:
        """Return the firmware_version."""
        return self.data.get("firmwareVersion")

    @property
    def zwave_plus_version(self) -> int | None:
        """Return the zwave_plus_version."""
        return self.data.get("zwavePlusVersion")

    @property
    def zwave_plus_node_type(self) -> int | None:
        """Return the zwave_plus_node_type."""
        return self.data.get("zwavePlusNodeType")

    @property
    def zwave_plus_role_type(self) -> int | None:
        """Return the zwave_plus_role_type."""
        return self.data.get("zwavePlusRoleType")

    @property
    def name(self) -> str | None:
        """Return the name."""
        return self.data.get("name")

    @property
    def location(self) -> str | None:
        """Return the location."""
        return self.data.get("location")

    @property
    def device_config(self) -> DeviceConfig:
        """Return the device_config."""
        return self._device_config

    @property
    def label(self) -> str | None:
        """Return the label."""
        return self.data.get("label")

    @property
    def device_database_url(self) -> str | None:
        """Return the device database URL."""
        return self.data.get("deviceDatabaseUrl")

    @property
    def endpoint_count_is_dynamic(self) -> bool | None:
        """Return the endpoint_count_is_dynamic."""
        return self.data.get("endpointCountIsDynamic")

    @property
    def endpoints_have_identical_capabilities(self) -> bool | None:
        """Return the endpoints_have_identical_capabilities."""
        return self.data.get("endpointsHaveIdenticalCapabilities")

    @property
    def individual_endpoint_count(self) -> int | None:
        """Return the individual_endpoint_count."""
        return self.data.get("individualEndpointCount")

    @property
    def aggregated_endpoint_count(self) -> int | None:
        """Return the aggregated_endpoint_count."""
        return self.data.get("aggregatedEndpointCount")

    @property
    def interview_attempts(self) -> int | None:
        """Return the interview_attempts."""
        return self.data.get("interviewAttempts")

    @property
    def interview_stage(self) -> int | str | None:
        """Return the interview_stage."""
        return self.data.get("interviewStage")

    @property
    def in_interview(self) -> bool:
        """Return whether node is currently being interviewed."""
        return (
            not self.ready
            and not self.awaiting_manual_interview
            and self.interview_stage != INTERVIEW_FAILED
        )

    @property
    def awaiting_manual_interview(self) -> bool:
        """Return whether node requires a manual interview."""
        return self.interview_stage in (None, NOT_INTERVIEWED)

    @property
    def command_classes(self) -> list[CommandClassInfo]:
        """Return all CommandClasses supported on this node."""
        return self.endpoints[0].command_classes

    @property
    def statistics(self) -> NodeStatistics:
        """Return statistics property."""
        return self._statistics

    @property
    def firmware_update_progress(self) -> NodeFirmwareUpdateProgress | None:
        """Return firmware update progress."""
        return self._firmware_update_progress

    @property
    def highest_security_class(self) -> SecurityClass | None:
        """Return highest security class configured on the node."""
        if (security_class := self.data.get("highestSecurityClass")) is None:
            return None
        return SecurityClass(security_class)

    @property
    def is_controller_node(self) -> bool:
        """Return whether the node is a controller node."""
        return self.data["isControllerNode"]

    @property
    def keep_awake(self) -> bool:
        """Return whether the node is set to keep awake."""
        return self.data["keepAwake"]

    @property
    def last_seen(self) -> datetime | None:
        """Return when the node was last seen."""
        return self._last_seen

    @property
    def default_volume(self) -> int | float | None:
        """Return the default volume."""
        return self.data.get("defaultVolume")

    @property
    def default_transition_duration(self) -> int | float | None:
        """Return the default transition duration."""
        return self.data.get("defaultTransitionDuration")

    @property
    def protocol(self) -> Protocols | None:
        """Return the protocol used to communicate with this node."""
        if "protocol" in self.data:
            return Protocols(self.data["protocol"])
        return None

    def _update_endpoints(self, endpoints: list[EndpointDataType]) -> None:
        """Update the endpoints data."""
        new_endpoints_data = {endpoint["index"]: endpoint for endpoint in endpoints}
        new_endpoint_idxs = set(new_endpoints_data)
        stale_endpoint_idxs = set(self.endpoints) - new_endpoint_idxs

        # Remove stale endpoints
        for endpoint_idx in stale_endpoint_idxs:
            self.endpoints.pop(endpoint_idx)

        # Add new endpoints or update existing ones
        for endpoint_idx in new_endpoint_idxs:
            endpoint = new_endpoints_data[endpoint_idx]
            values = {
                value_id: value
                for value_id, value in self.values.items()
                if self.index == value.endpoint
            }
            if endpoint_idx in self.endpoints:
                self.endpoints[endpoint_idx].update(endpoint, values)
            else:
                self.endpoints[endpoint_idx] = Endpoint(
                    self.client, self, endpoint, values
                )

    def _update_values(self, values: list[ValueDataType]) -> None:
        """Update the values data."""
        new_values_data = {
            _get_value_id_str_from_dict(self, val): val for val in values
        }
        new_value_ids = set(new_values_data)
        stale_value_ids = set(self.values) - new_value_ids

        # Remove stale values
        for value_id in stale_value_ids:
            self.values.pop(value_id)

        # Updating existing values and populate new values. Preserve value order if
        # initializing values for the node for the first time by using the key order
        # which is deterministic
        for value_id in (
            new_value_ids - stale_value_ids
            if stale_value_ids
            else list(new_values_data)
        ):
            val = new_values_data[value_id]
            try:
                if value_id in self.values:
                    self.values[value_id].update(val)
                else:
                    self.values[value_id] = self._init_value(val)
            except UnparseableValue:
                # If we can't parse the value, don't store it
                pass

    def update(self, data: NodeDataType) -> None:
        """Update the internal state data."""
        self.data = copy.deepcopy(data)
        self._device_config = DeviceConfig(self.data.get("deviceConfig", {}))
        if (device_class := self.data.get("deviceClass")) is None:
            self._device_class = None
        else:
            self._device_class = DeviceClass(device_class)

        self._statistics = NodeStatistics(
            self.client, self.data.get("statistics", DEFAULT_NODE_STATISTICS)
        )
        if last_seen := data.get("lastSeen"):
            self._last_seen = datetime.fromisoformat(last_seen)
        if not self._statistics.last_seen and self.last_seen:
            self._statistics.last_seen = self.last_seen
            self._statistics.data["lastSeen"] = self.last_seen.isoformat()

        self._update_values(self.data.pop("values"))
        self._update_endpoints(self.data.pop("endpoints"))

    def get_command_class_values(
        self, command_class: CommandClass, endpoint: int | None = None
    ) -> dict[str, ConfigurationValue | Value]:
        """Return all values for a given command class."""
        return {
            value_id: value
            for value_id, value in self.values.items()
            if value.command_class == command_class
            and (endpoint is None or value.endpoint == endpoint)
        }

    def get_configuration_values(self) -> dict[str, ConfigurationValue]:
        """Return all configuration values for a node."""
        return cast(
            dict[str, ConfigurationValue],
            self.get_command_class_values(CommandClass.CONFIGURATION),
        )

    def receive_event(self, event: Event) -> None:
        """Receive an event."""
        NODE_EVENT_MODEL_MAP[event.type].from_dict(event.data)

        self._handle_event_protocol(event)
        event.data["node"] = self

        self.emit(event.type, event.data)

    async def async_send_command(
        self,
        cmd: str,
        require_schema: int | None = None,
        wait_for_result: bool | None = None,
        **cmd_kwargs: Any,
    ) -> dict[str, Any] | None:
        """
        Send a node command. For internal use only.

        If wait_for_result is not None, it will take precedence, otherwise we will
        decide to wait or not based on the node status.
        """
        kwargs = {}
        message = {"command": f"node.{cmd}", "nodeId": self.node_id, **cmd_kwargs}
        if require_schema is not None:
            kwargs["require_schema"] = require_schema

        if wait_for_result:
            result = await self.client.async_send_command(message, **kwargs)
            return result
        if wait_for_result is None and self.status not in (NodeStatus.ASLEEP,):
            result_task = asyncio.create_task(
                self.client.async_send_command(message, **kwargs)
            )
            status_task = asyncio.create_task(self.status_event.wait())
            await asyncio.wait(
                [result_task, status_task],
                return_when=asyncio.FIRST_COMPLETED,
            )
            status_task.cancel()
            if self.status_event.is_set() and not result_task.done():
                result_task.cancel()
                return None
            return result_task.result()

        await self.client.async_send_command_no_wait(message, **kwargs)
        return None

    async def async_set_value(
        self,
        val: Value | str,
        new_value: Any,
        options: dict | None = None,
        wait_for_result: bool | None = None,
    ) -> SetValueResult | None:
        """Send setValue command to Node for given value (or value_id)."""
        # a value may be specified as value_id or the value itself
        if not isinstance(val, Value):
            if val not in self.values:
                raise NotFoundError(f"Value {val} not found on node {self}")
            val = self.values[val]

        if val.metadata.writeable is False:
            raise UnwriteableValue

        cmd_args = {
            "valueId": _get_value_id_dict_from_value_data(val.data),
            "value": new_value,
        }
        if options:
            option = next(
                (
                    option
                    for option in options
                    if option not in val.metadata.value_change_options
                ),
                None,
            )
            if option is not None:
                raise NotFoundError(
                    f"Option {option} not found on value {val} on node {self}"
                )
            cmd_args["options"] = options

        # the value object needs to be send to the server
        result = await self.async_send_command(
            "set_value", **cmd_args, require_schema=29, wait_for_result=wait_for_result
        )

        if result is None:
            return None

        return SetValueResult(result["result"])

    async def async_refresh_info(self) -> None:
        """Send refreshInfo command to Node."""
        await self.async_send_command("refresh_info", wait_for_result=False)

    async def async_refresh_values(self) -> None:
        """Send refreshValues command to Node."""
        await self.async_send_command(
            "refresh_values", wait_for_result=False, require_schema=4
        )

    async def async_refresh_cc_values(self, command_class: CommandClass) -> None:
        """Send refreshCCValues command to Node."""
        await self.async_send_command(
            "refresh_cc_values",
            commandClass=command_class,
            wait_for_result=False,
            require_schema=4,
        )

    async def async_get_defined_value_ids(self) -> list[Value]:
        """Send getDefinedValueIDs command to Node."""
        data = await self.async_send_command(
            "get_defined_value_ids", wait_for_result=True
        )
        assert data
        return [
            self._init_value(cast(ValueDataType, value_id))
            for value_id in data["valueIds"]
        ]

    async def async_get_value_metadata(self, val: Value | str) -> ValueMetadata:
        """Send getValueMetadata command to Node."""
        # a value may be specified as value_id or the value itself
        if not isinstance(val, Value):
            val = self.values[val]
        # the value object needs to be send to the server
        data = await self.async_send_command(
            "get_value_metadata",
            valueId=_get_value_id_dict_from_value_data(val.data),
            wait_for_result=True,
        )
        return ValueMetadata(cast(MetaDataType, data))

    async def async_get_firmware_update_capabilities(
        self,
    ) -> NodeFirmwareUpdateCapabilities:
        """Send getFirmwareUpdateCapabilities command to Node."""
        data = await self.async_send_command(
            "get_firmware_update_capabilities",
            require_schema=7,
            wait_for_result=True,
        )
        assert data
        return NodeFirmwareUpdateCapabilities(
            cast(NodeFirmwareUpdateCapabilitiesDataType, data["capabilities"])
        )

    async def async_get_firmware_update_capabilities_cached(
        self,
    ) -> NodeFirmwareUpdateCapabilities:
        """Send getFirmwareUpdateCapabilitiesCached command to Node."""
        data = await self.async_send_command(
            "get_firmware_update_capabilities_cached",
            require_schema=21,
            wait_for_result=True,
        )
        assert data
        return NodeFirmwareUpdateCapabilities(
            cast(NodeFirmwareUpdateCapabilitiesDataType, data["capabilities"])
        )

    async def async_abort_firmware_update(self) -> None:
        """Send abortFirmwareUpdate command to Node."""
        await self.async_send_command("abort_firmware_update", wait_for_result=False)

    async def async_poll_value(self, val: Value | str) -> None:
        """Send pollValue command to Node for given value (or value_id)."""
        # a value may be specified as value_id or the value itself
        if not isinstance(val, Value):
            val = self.values[val]
        await self.async_send_command(
            "poll_value",
            valueId=_get_value_id_dict_from_value_data(val.data),
            require_schema=1,
        )

    async def async_ping(self) -> bool:
        """Send ping command to Node."""
        data = (
            await self.async_send_command(
                "ping", require_schema=5, wait_for_result=True
            )
            or {}
        )
        return cast(bool, data.get("responded", False))

    async def async_invoke_cc_api(
        self,
        command_class: CommandClass,
        method_name: str,
        *args: Any,
        wait_for_result: bool | None = None,
    ) -> Any:
        """Call endpoint.invoke_cc_api command."""
        return await self.endpoints[0].async_invoke_cc_api(
            command_class, method_name, *args, wait_for_result=wait_for_result
        )

    async def async_supports_cc_api(self, command_class: CommandClass) -> bool:
        """Call endpoint.supports_cc_api command."""
        return await self.endpoints[0].async_supports_cc_api(command_class)

    async def async_supports_cc(self, command_class: CommandClass) -> bool:
        """Call endpoint.supports_cc command."""
        return await self.endpoints[0].async_supports_cc(command_class)

    async def async_controls_cc(self, command_class: CommandClass) -> bool:
        """Call endpoint.controls_cc command."""
        return await self.endpoints[0].async_controls_cc(command_class)

    async def async_is_cc_secure(self, command_class: CommandClass) -> bool:
        """Call endpoint.is_cc_secure command."""
        return await self.endpoints[0].async_is_cc_secure(command_class)

    async def async_get_cc_version(self, command_class: CommandClass) -> bool:
        """Call endpoint.get_cc_version command."""
        return await self.endpoints[0].async_get_cc_version(command_class)

    async def async_get_node_unsafe(self) -> NodeDataType:
        """Call endpoint.get_node_unsafe command."""
        return await self.endpoints[0].async_get_node_unsafe()

    async def async_has_security_class(
        self, security_class: SecurityClass
    ) -> bool | None:
        """Return whether node has the given security class."""
        data = await self.async_send_command(
            "has_security_class",
            securityClass=security_class,
            require_schema=8,
            wait_for_result=True,
        )
        if data and (has_security_class := data.get("hasSecurityClass")) is not None:
            return cast(bool, has_security_class)
        return None

    async def async_get_highest_security_class(self) -> SecurityClass | None:
        """Get the highest security class that a node supports."""
        data = await self.async_send_command(
            "get_highest_security_class", require_schema=8, wait_for_result=True
        )
        if data and (security_class := data.get("highestSecurityClass")) is not None:
            return SecurityClass(security_class)
        return None

    async def async_test_power_level(
        self, test_node: Node, power_level: PowerLevel, test_frame_count: int
    ) -> int:
        """Send testPowerLevel command to Node."""
        data = await self.async_send_command(
            "test_powerlevel",
            testNodeId=test_node.node_id,
            powerlevel=power_level,
            testFrameCount=test_frame_count,
            require_schema=13,
            wait_for_result=True,
        )
        assert data
        return cast(int, data["framesAcked"])

    async def async_check_lifeline_health(
        self, rounds: int | None = None
    ) -> LifelineHealthCheckSummary:
        """Send checkLifelineHealth command to Node."""
        kwargs = {}
        if rounds is not None:
            kwargs["rounds"] = rounds
        data = await self.async_send_command(
            "check_lifeline_health",
            require_schema=13,
            wait_for_result=True,
            **kwargs,
        )
        assert data
        return LifelineHealthCheckSummary(data["summary"])

    async def async_check_route_health(
        self, target_node: Node, rounds: int | None = None
    ) -> RouteHealthCheckSummary:
        """Send checkRouteHealth command to Node."""
        kwargs = {"targetNodeId": target_node.node_id}
        if rounds is not None:
            kwargs["rounds"] = rounds
        data = await self.async_send_command(
            "check_route_health",
            require_schema=13,
            wait_for_result=True,
            **kwargs,
        )
        assert data
        return RouteHealthCheckSummary(data["summary"])

    async def async_get_state(self) -> NodeDataType:
        """Get node state."""
        data = await self.async_send_command(
            "get_state", require_schema=14, wait_for_result=True
        )
        assert data
        return cast(NodeDataType, data["state"])

    async def async_set_name(
        self, name: str, update_cc: bool = True, wait_for_result: bool | None = None
    ) -> None:
        """Set node name."""
        # If we may not potentially update the name CC, we should just wait for the
        # result because the change is local to the driver
        if not update_cc:
            wait_for_result = True
        await self.async_send_command(
            "set_name",
            name=name,
            updateCC=update_cc,
            wait_for_result=wait_for_result,
            require_schema=14,
        )
        self.data["name"] = name

    async def async_set_location(
        self,
        location: str,
        update_cc: bool = True,
        wait_for_result: bool | None = None,
    ) -> None:
        """Set node location."""
        # If we may not potentially update the location CC, we should just wait for the
        # result because the change is local to the driver
        if not update_cc:
            wait_for_result = True
        await self.async_send_command(
            "set_location",
            location=location,
            updateCC=update_cc,
            wait_for_result=wait_for_result,
            require_schema=14,
        )
        self.data["location"] = location

    async def async_is_firmware_update_in_progress(self) -> bool:
        """
        Send isFirmwareUpdateInProgress command to Node.

        If `True`, a firmware update for this node is in progress.
        """
        data = await self.async_send_command(
            "is_firmware_update_in_progress", require_schema=21, wait_for_result=True
        )
        assert data
        return cast(bool, data["progress"])

    async def async_set_keep_awake(self, keep_awake: bool) -> None:
        """Set node keep awake state."""
        await self.async_send_command(
            "set_keep_awake",
            keepAwake=keep_awake,
            wait_for_result=True,
            require_schema=14,
        )
        self.data["keepAwake"] = keep_awake

    async def async_interview(self) -> None:
        """Interview node."""
        await self.async_send_command(
            "interview",
            wait_for_result=False,
            require_schema=22,
        )

    async def async_get_value_timestamp(self, val: Value | str) -> int:
        """Send getValueTimestamp command to Node for given value (or value_id)."""
        # a value may be specified as value_id or the value itself
        if not isinstance(val, Value):
            val = self.values[val]
        data = await self.async_send_command(
            "get_value_timestamp",
            valueId=_get_value_id_dict_from_value_data(val.data),
            require_schema=27,
            wait_for_result=True,
        )
        assert data
        return cast(int, data["timestamp"])

    async def async_manually_idle_notification_value(self, val: Value | str) -> None:
        """Send manuallyIdleNotificationValue cmd to Node for value (or value_id)."""
        # a value may be specified as value_id or the value itself
        if not isinstance(val, Value):
            val = self.values[val]
        if val.command_class != CommandClass.NOTIFICATION:
            raise ValueError(
                "Value must be of CommandClass.NOTIFICATION to manually idle it"
            )
        await self.async_send_command(
            "manually_idle_notification_value",
            valueId=_get_value_id_dict_from_value_data(val.data),
            require_schema=28,
            wait_for_result=False,
        )

    async def async_set_date_and_time(
        self, datetime_: datetime | None = None, wait_for_result: bool | None = None
    ) -> bool | None:
        """Send setDateAndTime command to Node."""
        args = {}
        if datetime_:
            args["date"] = datetime_.isoformat()
        data = await self.async_send_command(
            "set_date_and_time",
            **args,
            require_schema=28,
            wait_for_result=wait_for_result,
        )
        if data:
            return cast(bool, data["success"])
        return None

    async def async_get_date_and_time(self) -> DateAndTime:
        """Send getDateAndTime command to Node."""
        data = await self.async_send_command(
            "get_date_and_time",
            require_schema=31,
            wait_for_result=True,
        )
        assert data
        return DateAndTime(data["dateAndTime"])

    async def async_is_health_check_in_progress(self) -> bool:
        """Send isHealthCheckInProgress command to Node."""
        data = await self.async_send_command(
            "is_health_check_in_progress",
            require_schema=31,
            wait_for_result=True,
        )
        assert data
        return cast(bool, data["progress"])

    async def async_abort_health_check(self) -> None:
        """Send abortHealthCheck command to Node."""
        await self.async_send_command(
            "abort_health_check",
            require_schema=31,
            wait_for_result=False,
        )

    async def async_set_default_volume(
        self, default_volume: int | float | None
    ) -> None:
        """Send setDefaultVolume command to Node."""
        cmd_kwargs = {}
        self.data["defaultVolume"] = default_volume
        if default_volume is not None:
            cmd_kwargs["defaultVolume"] = default_volume
        await self.async_send_command(
            "set_default_volume",
            require_schema=31,
            wait_for_result=None,
            **cmd_kwargs,
        )

    async def async_set_default_transition_duration(
        self, default_duration_transition: int | float | None
    ) -> None:
        """Send setDefaultTransitionDuration command to Node."""
        cmd_kwargs = {}
        self.data["defaultTransitionDuration"] = default_duration_transition
        if default_duration_transition is not None:
            cmd_kwargs["defaultTransitionDuration"] = default_duration_transition
        await self.async_send_command(
            "set_default_transition_duration",
            require_schema=31,
            wait_for_result=None,
            **cmd_kwargs,
        )

    async def async_has_device_config_changed(self) -> bool | None:
        """Send hasDeviceConfigChanged command to Node."""
        data = await self.async_send_command(
            "has_device_config_changed",
            require_schema=31,
            wait_for_result=True,
        )
        if data and (changed := data.get("changed")) is not None:
            return cast(bool, changed)
        return None

    async def async_set_raw_config_parameter_value(
        self,
        new_value: int,
        property_: int | str,
        property_key: int | None = None,
        value_size: Literal[1, 2, 4] | None = None,
        value_format: ConfigurationValueFormat | None = None,
    ) -> SetConfigParameterResult:
        """Send setRawConfigParameterValue."""
        return await self.endpoints[0].async_set_raw_config_parameter_value(
            new_value, property_, property_key, value_size, value_format
        )

    async def async_get_raw_config_parameter_value(
        self,
        property_: int,
        property_key: int | None = None,
        allow_unexpected_response: bool | None = None,
    ) -> int:
        """Call getRawConfigParameterValue."""
        return await self.endpoints[0].async_get_raw_config_parameter_value(
            property_, property_key, allow_unexpected_response
        )

    def handle_test_powerlevel_progress(self, event: Event) -> None:
        """Process a test power level progress event."""
        event.data["test_power_level_progress"] = TestPowerLevelProgress(
            event.data["acknowledged"], event.data["total"]
        )

    def handle_check_lifeline_health_progress(self, event: Event) -> None:
        """Process a check lifeline health progress event."""
        event.data["check_lifeline_health_progress"] = CheckHealthProgress(
            event.data["rounds"],
            event.data["totalRounds"],
            event.data["lastRating"],
            LifelineHealthCheckResult(event.data["lastResult"]),
        )

    def handle_check_route_health_progress(self, event: Event) -> None:
        """Process a check route health progress event."""
        event.data["check_route_health_progress"] = CheckHealthProgress(
            event.data["rounds"],
            event.data["totalRounds"],
            event.data["lastRating"],
            RouteHealthCheckResult(event.data["lastResult"]),
        )

    def handle_wake_up(self, event: Event) -> None:
        """Process a node wake up event."""
        # pylint: disable=unused-argument
        self.status_event.clear()
        self.data["status"] = NodeStatus.AWAKE

    def handle_sleep(self, event: Event) -> None:
        """Process a node sleep event."""
        # pylint: disable=unused-argument
        self.status_event.set()
        self.data["status"] = NodeStatus.ASLEEP

    def handle_dead(self, event: Event) -> None:
        """Process a node dead event."""
        # pylint: disable=unused-argument
        self.status_event.set()
        self.data["status"] = NodeStatus.DEAD

    def handle_alive(self, event: Event) -> None:
        """Process a node alive event."""
        # pylint: disable=unused-argument
        self.status_event.clear()
        self.data["status"] = NodeStatus.ALIVE

    def handle_interview_started(self, event: Event) -> None:
        """Process a node interview started event."""
        # pylint: disable=unused-argument
        self.data["ready"] = False
        self.data["interviewStage"] = None

    def handle_interview_stage_completed(self, event: Event) -> None:
        """Process a node interview stage completed event."""
        self.data["interviewStage"] = event.data["stageName"]

    def handle_interview_failed(self, event: Event) -> None:
        """Process a node interview failed event."""
        # pylint: disable=unused-argument
        self.data["interviewStage"] = INTERVIEW_FAILED

    def handle_interview_completed(self, event: Event) -> None:
        """Process a node interview completed event."""
        # pylint: disable=unused-argument
        self.data["ready"] = True

    def handle_ready(self, event: Event) -> None:
        """Process a node ready event."""
        # the event contains a full dump of the node
        self.update(event.data["nodeState"])

    def handle_value_added(self, event: Event) -> None:
        """Process a node value added event."""
        self.handle_value_updated(event)

    def handle_value_updated(self, event: Event) -> None:
        """Process a node value updated event."""
        evt_val_data: ValueDataType = event.data["args"]
        value_id = _get_value_id_str_from_dict(self, evt_val_data)
        value = self.values.get(value_id)
        if value is None:
            value = self._init_value(evt_val_data)
            self.values[value.value_id] = event.data["value"] = value
        else:
            value.receive_event(event)
            event.data["value"] = value

    def handle_value_removed(self, event: Event) -> None:
        """Process a node value removed event."""
        value_id = _get_value_id_str_from_dict(self, event.data["args"])
        event.data["value"] = self.values.pop(value_id)

    def handle_value_notification(self, event: Event) -> None:
        """Process a node value notification event."""
        # if value is found, use value data as base and update what is provided
        # in the event, otherwise use the event data
        event_data = event.data["args"]
        if value := self.values.get(_get_value_id_str_from_dict(self, event_data)):
            value_notification = ValueNotification(
                self, cast(ValueDataType, dict(value.data))
            )
            value_notification.update(event_data)
        else:
            value_notification = ValueNotification(self, event_data)

        event.data["value_notification"] = value_notification

    def handle_metadata_updated(self, event: Event) -> None:
        """Process a node metadata updated event."""
        # handle metadata updated as value updated (as its a value object with
        # included metadata)
        self.handle_value_updated(event)

    def handle_notification(self, event: Event) -> None:
        """Process a node notification event."""
        match command_class := CommandClass(event.data["ccId"]):
            case CommandClass.NOTIFICATION:
                event.data["notification"] = NotificationNotification(
                    self, cast(NotificationNotificationDataType, event.data)
                )
            case CommandClass.SWITCH_MULTILEVEL:
                event.data["notification"] = MultilevelSwitchNotification(
                    self, cast(MultilevelSwitchNotificationDataType, event.data)
                )
            case CommandClass.ENTRY_CONTROL:
                event.data["notification"] = EntryControlNotification(
                    self, cast(EntryControlNotificationDataType, event.data)
                )
            case CommandClass.POWERLEVEL:
                event.data["notification"] = PowerLevelNotification(
                    self, cast(PowerLevelNotificationDataType, event.data)
                )
            case _:
                _LOGGER.info(
                    "Unhandled notification command class: %s", command_class.name
                )

    def handle_firmware_update_progress(self, event: Event) -> None:
        """Process a node firmware update progress event."""
        self._firmware_update_progress = event.data["firmware_update_progress"] = (
            NodeFirmwareUpdateProgress(
                self, cast(NodeFirmwareUpdateProgressDataType, event.data["progress"])
            )
        )

    def handle_firmware_update_finished(self, event: Event) -> None:
        """Process a node firmware update finished event."""
        self._firmware_update_progress = None
        event.data["firmware_update_finished"] = NodeFirmwareUpdateResult(
            self, cast(NodeFirmwareUpdateResultDataType, event.data["result"])
        )

    def handle_statistics_updated(self, event: Event) -> None:
        """Process a statistics updated event."""
        self.data["statistics"] = statistics = event.data["statistics"]
        event.data["statistics_updated"] = self._statistics = NodeStatistics(
            self.client, statistics
        )
        if self._statistics.last_seen:
            self._last_seen = self._statistics.last_seen



"""Data model for a Z-Wave JS node."""

from __future__ import annotations

from typing import TypedDict

from ..device_class import DeviceClassDataType
from ..device_config import DeviceConfigDataType
from ..endpoint import EndpointDataType
from ..value import ValueDataType
from .statistics import NodeStatisticsDataType


class FoundNodeDataType(TypedDict, total=False):
    """Represent a found node data dict type."""

    nodeId: int
    deviceClass: DeviceClassDataType
    supportedCCs: list[int]
    controlledCCs: list[int]


class NodeDataType(TypedDict, total=False):
    """Represent a node data dict type."""

    nodeId: int  # required
    index: int  # required
    deviceClass: DeviceClassDataType | None
    installerIcon: int
    userIcon: int
    name: str
    location: str
    status: int  # 0-4  # required
    zwavePlusVersion: int
    zwavePlusNodeType: int
    zwavePlusRoleType: int
    isListening: bool
    isFrequentListening: bool | str
    isRouting: bool
    maxDataRate: int
    supportedDataRates: list[int]
    isSecure: bool
    supportsBeaming: bool
    supportsSecurity: bool
    protocolVersion: int
    firmwareVersion: str
    manufacturerId: int
    productId: int
    productType: int
    deviceConfig: DeviceConfigDataType
    deviceDatabaseUrl: str
    keepAwake: bool
    ready: bool
    label: str
    endpoints: list[EndpointDataType]
    endpointCountIsDynamic: bool
    endpointsHaveIdenticalCapabilities: bool
    individualEndpointCount: int
    aggregatedEndpointCount: int
    interviewAttempts: int
    interviewStage: int | str | None
    values: list[ValueDataType]
    statistics: NodeStatisticsDataType
    highestSecurityClass: int
    isControllerNode: bool
    lastSeen: str
    defaultVolume: int | float | None
    defaultTransitionDuration: int | float | None
    protocol: int
    sdkVersion: str | None



"""Test the config manager."""

from zwave_js_server.client import Client
from zwave_js_server.model.config_manager import ConfigManager
from zwave_js_server.model.device_config import DeviceConfigDataType

from ..common import MockCommandProtocol


async def test_lookup_device(
    client: Client,
    mock_command: MockCommandProtocol,
    device_config: DeviceConfigDataType,
) -> None:
    """Test the lookup_device command."""
    # Test successful lookup
    mock_command(
        {
            "command": "config_manager.lookup_device",
            "manufacturerId": 0x1234,
            "productType": 0x5678,
            "productId": 0x9ABC,
        },
        {"config": device_config},
    )

    config_manager = ConfigManager(client)
    result = await config_manager.lookup_device(0x1234, 0x5678, 0x9ABC)

    assert result is not None
    assert result.manufacturer == "Test Manufacturer"
    assert result.manufacturer_id == "0x1234"
    assert result.label == "Test Device"
    assert result.description == "Test Device Description"
    assert len(result.devices) == 1
    assert result.devices[0].product_type == "0x5678"
    assert result.devices[0].product_id == "0x9ABC"
    assert result.firmware_version.min == "1.0"
    assert result.firmware_version.max == "2.0"

    # Test lookup with firmware version
    mock_command(
        {
            "command": "config_manager.lookup_device",
            "manufacturerId": 0x1234,
            "productType": 0x5678,
            "productId": 0x9ABC,
            "firmwareVersion": "1.5",
        },
        {"config": device_config},
    )

    result = await config_manager.lookup_device(0x1234, 0x5678, 0x9ABC, "1.5")

    assert result is not None
    assert result.manufacturer == "Test Manufacturer"

    assert result.to_dict() == device_config

    # Test device not found
    mock_command(
        {
            "command": "config_manager.lookup_device",
            "manufacturerId": 0x4321,
            "productType": 0x8765,
            "productId": 0xCBA9,
        },
        {"config": None},
    )

    result = await config_manager.lookup_device(0x4321, 0x8765, 0xCBA9)

    assert result is None



"""Test the server version helper."""

from unittest.mock import call

from zwave_js_server.version import get_server_version


async def test_get_server_version(client_session, ws_client, url, version_data):
    """Test the get server version helper."""
    ws_client.receive_json.return_value = version_data

    version_info = await get_server_version(url, client_session)

    assert client_session.ws_connect.called
    assert client_session.ws_connect.call_args == call(url)
    assert version_info.driver_version == version_data["driverVersion"]
    assert version_info.server_version == version_data["serverVersion"]
    assert version_info.home_id == version_data["homeId"]
    assert version_info.min_schema_version == version_data["minSchemaVersion"]
    assert version_info.max_schema_version == version_data["maxSchemaVersion"]
    assert ws_client.close.called


async def test_missing_server_schema_version(
    client_session, ws_client, url, version_data
):
    """Test missing schema version processed as schema version 0."""
    del version_data["minSchemaVersion"]
    del version_data["maxSchemaVersion"]
    ws_client.receive_json.return_value = version_data
    version_info = await get_server_version(url, client_session)
    assert version_info.min_schema_version == 0
    assert version_info.max_schema_version == 0
    assert ws_client.close.called



"""Data model for a Z-Wave JS controller."""

from __future__ import annotations

from typing import TypedDict

from .statistics import ControllerStatisticsDataType


class ControllerDataType(TypedDict, total=False):
    """Represent a controller data dict type."""

    sdkVersion: str
    type: int
    homeId: int
    ownNodeId: int
    isPrimary: bool
    isSUC: bool
    nodeType: int
    isUsingHomeIdFromOtherNetwork: bool
    isSISPresent: bool
    wasRealPrimary: bool
    firmwareVersion: str
    manufacturerId: int
    productType: int
    productId: int
    supportedFunctionTypes: list[int]
    sucNodeId: int
    supportsTimers: bool
    isRebuildingRoutes: bool
    statistics: ControllerStatisticsDataType
    inclusionState: int
    rfRegion: int
    status: int
    rebuildRoutesProgress: dict[str, str]
    supportsLongRange: bool
